"""
Modified from https://github.com/kaijieshi7/Dynamic-convolution-Pytorch/blob/master/dynamic_conv.py
"""

import oneflow as flow
import oneflow.nn as nn
import oneflow.nn.functional as F


class attention2d(nn.Module):
    def __init__(self, in_planes, ratios, K, temperature):
        super(attention2d, self).__init__()
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        hidden_planes = int(in_planes * ratios)
        self.fc1 = nn.Conv2d(in_planes, hidden_planes, 1, bias=False)
        self.fc2 = nn.Conv2d(hidden_planes, K, 1, bias=True)
        self.temperature = temperature

    def update_temperature(self, annealing=1):
        assert self.temperature >= 1
        if self.temperature != 1:
            self.temperature -= annealing
            print("Change temperature to:", str(self.temperature))

    def forward(self, x):
        x = self.avgpool(x)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.fc2(x).view(x.size(0), -1)
        return F.softmax(x / self.temperature, 1)


class Dynamic_conv2d(nn.Module):
    """`Dynamic Convolution: Attention over Convolution Kernels <https://arxiv.org/abs/1912.03458>`_"""

    def __init__(
        self,
        in_planes,
        out_planes,
        kernel_size,
        ratio=0.25,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        K=4,
        temperature=34,
    ):
        super(Dynamic_conv2d, self).__init__()
        assert in_planes % groups == 0
        self.in_planes = in_planes
        self.out_planes = out_planes
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.bias = bias
        self.K = K
        self.attention = attention2d(in_planes, ratio, K, temperature)

        self.weight = nn.Parameter(
            flow.randn(K, out_planes, in_planes // groups, kernel_size, kernel_size),
            requires_grad=True,
        )
        if bias:
            self.bias = nn.Parameter(flow.zeros(K, out_planes))
        else:
            self.bias = None

    def update_temperature(self):
        self.attention.update_temperature()

    def forward(self, x):
        softmax_attention = self.attention(x)
        batch_size, in_planes, height, width = x.size()
        x = x.view(1, -1, height, width)
        weight = self.weight.view(self.K, -1)
        aggregate_weight = (softmax_attention @ weight).view(
            batch_size * self.out_planes,
            self.in_planes // self.groups,
            self.kernel_size,
            self.kernel_size,
        )
        if self.bias is not None:
            aggregate_bias = (softmax_attention @ self.bias).view(
                batch_size * self.out_planes
            )
            output = F.conv2d(
                x,
                weight=aggregate_weight,
                bias=aggregate_bias,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.groups * batch_size,
                channel_pos="channels_first",
            )
        else:
            output = F.conv2d(
                x,
                weight=aggregate_weight,
                bias=None,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.groups * batch_size,
                channel_pos="channels_first",
            )
        output = output.view(
            batch_size, self.out_planes, output.size(-2), output.size(-1)
        )
        return output
